@@ -15,6 +15,7 @@
 #    along with this program; if not, write to the Free Software
 #    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 
+import datetime
 import hmac
 import socket
 import json
@@ -100,8 +101,10 @@ class IMAPServer(object):
         self.sslversion = repos.getsslversion()
         self.starttls = repos.getstarttls()
 
-        if self.tlslevel is not "tls_compat" and self.sslversion is None:
-            raise Exception("When 'tls_version' is not 'tls_compat' "
+        if self.usessl \
+           and self.tlslevel is not "tls_compat" \
+           and self.sslversion is None:
+            raise Exception("When 'tls_level' is not 'tls_compat' "
                 "the 'ssl_version' must be set explicitly.")
 
         self.oauth2_refresh_token = repos.getoauth2_refresh_token()
@@ -109,6 +112,7 @@ class IMAPServer(object):
         self.oauth2_client_id = repos.getoauth2_client_id()
         self.oauth2_client_secret = repos.getoauth2_client_secret()
         self.oauth2_request_url = repos.getoauth2_request_url()
+        self.oauth2_access_token_expires_at = None
 
         self.delim = None
         self.root = None
@@ -175,8 +179,7 @@ class IMAPServer(object):
 
         # get 1) configured password first 2) fall back to asking via UI
         self.password = self.repos.getpassword() or \
-            self.ui.getpass(self.repos.getname(), self.config,
-                self.passworderror)
+            self.ui.getpass(self.username, self.config, self.passworderror)
         self.passworderror = None
         return self.password
 
@@ -200,6 +203,11 @@ class IMAPServer(object):
           http://tools.ietf.org/html/rfc4616"""
 
         authc = self.username
+        if not authc:
+            raise OfflineImapError("No username provided for '%s'"
+                                    % self.repos.getname(),
+                                   OfflineImapError.ERROR.REPO)
+
         passwd = self.__getpassword()
         authz = b''
         if self.user_identity != None:
@@ -213,6 +221,12 @@ class IMAPServer(object):
         return retval
 
     def __xoauth2handler(self, response):
+        now = datetime.datetime.now()
+        if self.oauth2_access_token_expires_at \
+                and self.oauth2_access_token_expires_at < now:
+            self.oauth2_access_token = None
+            self.ui.debug('imap', 'xoauth2handler: oauth2_access_token expired')
+
         if self.oauth2_access_token is None:
             if self.oauth2_request_url is None:
                 raise OfflineImapError("No remote oauth2_request_url for "
@@ -250,9 +264,13 @@ class IMAPServer(object):
                 raise OfflineImapError("xoauth2handler got: %s"% resp,
                     OfflineImapError.ERROR.REPO)
             self.oauth2_access_token = resp['access_token']
+            if u'expires_in' in resp:
+                self.oauth2_access_token_expires_at = now + datetime.timedelta(
+                    seconds=resp['expires_in']/2
+                )
 
-        self.ui.debug('imap', 'xoauth2handler: access_token "%s"'%
-            self.oauth2_access_token)
+        self.ui.debug('imap', 'xoauth2handler: access_token "%s expires %s"'% (
+            self.oauth2_access_token, self.oauth2_access_token_expires_at))
         auth_string = 'user=%s\1auth=Bearer %s\1\1'% (
             self.username, self.oauth2_access_token)
         #auth_string = base64.b64encode(auth_string)
@@ -274,6 +292,13 @@ class IMAPServer(object):
             if not self.gss_vc.complete:
                 response = self.gss_vc.step(token)
                 return response if response else ""
+            elif token is None:
+                # uh... context is complete, so there's no negotiation we can
+                # do.  But we also don't have a token, so we can't send any
+                # kind of response.  Empirically, some (but not all) servers
+                # seem to put us in this state, and seem fine with getting no
+                # GSSAPI content in response, so give it to them.
+                return ""
 
             # Don't bother checking qop because we're over a TLS channel
             # already.  But hey, if some server started encrypting tomorrow,
